/****************************** Module Header ******************************\
* Module Name:    WorkerRole.cs
* Project:        MSDNArduinoListener
* 
* This source is subject to the Microsoft Public License.
* See http://www.microsoft.com/en-us/openness/licenses.aspx#MPL.
* All other rights reserved.
* 
* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
* EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
* WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
\*****************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Threading;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.Diagnostics;
using Microsoft.WindowsAzure.ServiceRuntime;
using Microsoft.WindowsAzure.Storage;
using Microsoft.ServiceBus.Messaging;
using System.Threading.Tasks;
using System.Net.Sockets;
using System.IO;
using Microsoft.ServiceBus;

namespace ArduinoListener
{
    public class WorkerRole : RoleEntryPoint
    {
        public override void Run()
        {
            Trace.WriteLine("Starting Application", "Information");

            var deviceEp = RoleEnvironment.CurrentRoleInstance.InstanceEndpoints["arduinoEP"];
            var messagingFactory = MessagingFactory.CreateFromConnectionString(RoleEnvironment.GetConfigurationSettingValue("serviceBusConnectionString"));
            
            SwitchServer.Run(deviceEp.IPEndpoint, messagingFactory);
        }

        public static class SwitchServer
        {
            static int timeout = 10000;
            static readonly byte[] OnFrame = { 0x01 };
            static readonly byte[] OffFrame = { 0x02 };

            public static void Run(IPEndPoint deviceEndPoint, MessagingFactory messagingFactory)
            {
                Trace.WriteLine("Starting Server", "Information");

                Task.WaitAll(new[]
                {
                    RunDeviceEndpoint(deviceEndPoint, messagingFactory)
                });
            }

            static async Task RunDeviceEndpoint(IPEndPoint deviceEP, MessagingFactory messagingFactory)
            {
                var deviceServer = new TcpListener(deviceEP);
                deviceServer.Start(10);

                Trace.WriteLine("Listening for connections", "Information");

                try
                {
                    do
                    {
                        TcpClient connection = await deviceServer.AcceptTcpClientAsync();
                        if (connection != null)
                        {
                            try
                            {
                                connection.NoDelay = true; 
                                connection.ReceiveTimeout = timeout;
                                NetworkStream deviceConnectionStream = connection.GetStream();
                                var readBuffer = new byte[64];

                                if (await deviceConnectionStream.ReadAsync(readBuffer, 0, 4) == 4)
                                {
                                    int deviceId = IPAddress.NetworkToHostOrder(BitConverter.ToInt32(readBuffer, 0));
                                    Trace.WriteLine("Received signal for device:" + deviceId, "Information");

                                    try
                                    {
                                        //Create device queue if it doesn't exist
                                        var namespaceManager = NamespaceManager.CreateFromConnectionString(RoleEnvironment.GetConfigurationSettingValue("serviceBusConnectionString"));
                                        if (!namespaceManager.QueueExists(string.Format("dev{0:X8}", deviceId)))
                                        {
                                            namespaceManager.CreateQueue(string.Format("dev{0:X8}", deviceId));
                                        }

                                        Console.WriteLine(string.Format("dev{0:X8}", deviceId), "Information");
                                        // set up the receiver for the per-device queue and sender for the response queue
                                        var deviceQueueReceiver = messagingFactory.CreateMessageReceiver(string.Format("dev{0:X8}", deviceId), ReceiveMode.PeekLock);

                                        Trace.WriteLine("Connected to Queue");

                                        do
                                        {
                                            // receive from the queue
                                            Trace.WriteLine("Reading the Queue", "Information");
                                            BrokeredMessage message = null;
                                            message = await deviceQueueReceiver.ReceiveAsync();

                                            try
                                            {
                                                if (message != null)
                                                {
                                                    Stream stream = message.GetBody<Stream>();
                                                    StreamReader reader = new StreamReader(stream);
                                                    string command = reader.ReadToEnd();
                                                    
                                                    if (command != null)
                                                    {
                                                        switch (command.ToUpperInvariant())
                                                        {
                                                            case "ON":
                                                                await deviceConnectionStream.WriteAsync(OnFrame, 0, OnFrame.Length);
                                                                await message.CompleteAsync();
                                                                break;
                                                            case "OFF":
                                                                await deviceConnectionStream.WriteAsync(OffFrame, 0, OffFrame.Length);
                                                                await message.CompleteAsync();
                                                                break;
                                                        }
                                                    }
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                Trace.TraceError(e.Message);
                                                deviceQueueReceiver.Close();
                                            }
                                        }
                                        while (true);
                                    }
                                    catch (Exception e)
                                    {
                                        Trace.TraceError(e.Message);
                                    }
                                    finally
                                    {
                                        connection.Close();
                                    }
                                }
                                else
                                {
                                    connection.Close();
                                }
                            }
                            catch (SocketException se)
                            {
                                // log
                                Trace.TraceError(se.Message);
                                connection.Close();
                            }
                            catch (IOException ioe)
                            {
                                Trace.TraceError(ioe.Message);
                                connection.Close();
                            }
                            catch (Exception e)
                            {
                                Trace.TraceError(e.Message);
                            }
                        }
                    } while (true);
                }
                catch (Exception e)
                {
                    Trace.TraceError(e.Message);
                }
                finally
                {
                    deviceServer.Stop();
                }
            }
        }
    }
}
