﻿/****************************** Module Header ******************************\
* Module Name:    Program.cs
* Project:        MSDNArduinoListener
* 
* This source is subject to the Microsoft Public License.
* See http://www.microsoft.com/en-us/openness/licenses.aspx#MPL.
* All other rights reserved.
* 
* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
* EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
* WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
\*****************************************************************************/

using Microsoft.ServiceBus.Messaging;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ConsoleListener
{
    class Program
    {
        static int timeout = 10000;
        static readonly byte[] OnFrame = { 0x01 };
        static readonly byte[] OffFrame = { 0x02 };

        static void Main(string[] args)
        {
            Console.WriteLine("Starting Application");
            Trace.WriteLine("Starting Application", "Information");

            String serviceBusConnectionString = "Endpoint=sb://msdnarduinoservicebus.servicebus.windows.net/;SharedSecretIssuer=owner;SharedSecretValue=qWjqKJFfJwB5JEkp5eI7ZfePYkrq6tnvFvn8X8hAIRc=";
            
            IPEndPoint deviceEp = CreateIPEndPoint("192.168.15.100:10100");
            var messagingFactory = MessagingFactory.CreateFromConnectionString(serviceBusConnectionString);

            RunServer(deviceEp, messagingFactory);
        }

        private static void RunServer(IPEndPoint deviceEndPoint, MessagingFactory messagingFactory)
        {
            Console.WriteLine("Starting Server");

            Task.WaitAll(new[] {RunDeviceEndpointAsync(deviceEndPoint, messagingFactory)});

        }

        private static async Task RunDeviceEndpointAsync(IPEndPoint deviceEP, MessagingFactory messagingFactory)
        {
            Console.WriteLine("Listening for connections");
            var deviceServer = new TcpListener(deviceEP);
            deviceServer.Start(10);

            try
            {
                do
                {
                    TcpClient connection = await deviceServer.AcceptTcpClientAsync();
                    if (connection != null)
                    {
                        try
                        {
                            connection.NoDelay = true; // flush writes immediately to the wire
                            connection.ReceiveTimeout = timeout;
                            NetworkStream deviceConnectionStream = connection.GetStream();
                            var readBuffer = new byte[64];
                                
                            if (await deviceConnectionStream.ReadAsync(readBuffer, 0, 4) == 4)
                            {
                                int deviceId = IPAddress.NetworkToHostOrder(BitConverter.ToInt32(readBuffer, 0));
                                Console.WriteLine("Received signal for device:" + deviceId);
                                Trace.WriteLine("Received signal for device:" + deviceId);

                                try
                                {
                                    //Create device queue if it doesn't exist
                                    var namespaceManager = NamespaceManager.CreateFromConnectionString(RoleEnvironment.GetConfigurationSettingValue("serviceBusConnectionString"));
                                    if (!namespaceManager.QueueExists(string.Format("dev{0:X8}", deviceId)))
                                    {
                                        namespaceManager.CreateQueue(string.Format("dev{0:X8}", deviceId));
                                    }

                                    Console.WriteLine(string.Format("dev{0:X8}", deviceId));
                                    // set up the receiver for the per-device queue and sender for the response queue
                                    var deviceQueueReceiver = messagingFactory.CreateMessageReceiver(string.Format("dev{0:X8}", deviceId), ReceiveMode.PeekLock);
                                        
                                    Console.WriteLine("Connected to Queue");

                                    do
                                    {
                                        // receive from the queue
                                        Console.WriteLine("Reading the Queue");
                                        Trace.WriteLine("Reading the Queue");
                                        BrokeredMessage message = null;
                                        message = await deviceQueueReceiver.ReceiveAsync();

                                        try 
                                        {
                                            if (message != null)
                                            {
                                                Stream stream = message.GetBody<Stream>();
                                                StreamReader reader = new StreamReader(stream);
                                                string command = reader.ReadToEnd();
                                                    
                                                if (command != null)
                                                {
                                                    switch (command.ToUpperInvariant())
                                                    {
                                                        case "ON":
                                                            await deviceConnectionStream.WriteAsync(OnFrame, 0, OnFrame.Length);
                                                            await message.CompleteAsync();
                                                            break;
                                                        case "OFF":
                                                            await deviceConnectionStream.WriteAsync(OffFrame, 0, OffFrame.Length);
                                                            await message.CompleteAsync();
                                                            break;
                                                    }
                                                }
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            Console.WriteLine(e.Message);
                                            deviceQueueReceiver.Close();
                                        }
                                    }
                                    while (true);
                                }
                                catch (Exception e)
                                {
                                    Console.WriteLine(e.Message);
                                }
                                finally
                                {
                                    connection.Close();
                                }
                            }
                            else
                            {
                                connection.Close();
                            }
                        }
                        catch (SocketException se)
                        {
                            // log
                            Console.WriteLine(se.Message);
                            connection.Close();
                        }
                        catch (IOException ioe)
                        {
                            Console.WriteLine(ioe.Message);
                            connection.Close();
                        }
                        catch (Exception e)
                        {
                            Console.WriteLine(e.Message);
                        }
                    }
                } while (true);
            }
            catch (Exception e)
            {
                Trace.TraceError(e.Message);
            }
            finally
            {
                deviceServer.Stop();
            }
        }
        
        private static IPEndPoint CreateIPEndPoint(string endPoint)
        {
            string[] ep = endPoint.Split(':');
            if (ep.Length != 2) throw new FormatException("Invalid endpoint format");
            IPAddress ip;
            if (!IPAddress.TryParse(ep[0], out ip))
            {
                throw new FormatException("Invalid ip-adress");
            }
            int port;
            if (!int.TryParse(ep[1], NumberStyles.None, NumberFormatInfo.CurrentInfo, out port))
            {
                throw new FormatException("Invalid port");
            }
            return new IPEndPoint(ip, port);
        }
    }
}
